////////////////////////////////////////////////////////////////////////////
//
//  parser.c
//
//	Simple Parser Source File
//    - Keywords as well as "Value-Codecs" can be defined
//    - optimized for 
//				-	RAM Usage (no input buffer needed, buffers needed are configurable)
//				- Speed (sliced token recognition)
//    - configurable in a wide range
//		- can be used for 
//			- simple command shells on a uC
//			- simple debug ui/command shell for applications
//			- configuring application data (value-codecs)
//
//  Strategy of Tokenizer : 
//  
//  With a list of all used tokens we take one bit per token which is set to one
//  at the beginning (we assume that all is possible :-)). At the 
//  beginning of parsing (first character) we scan every token's first char. 
//  Those which match get a one in their bit. At the next character we only 
//  check those having a one in their bit. And mark again ... and next character ..
//  When incorrect (not matching) characters occur, we reset the token recognition.
//	If only one possible token is left and length is correct we have found our token.
//  
//
//  Working principle of the Parser :
//
//  Each token is stored in a Buffer
//  The token found can be a treated as a Keyword, a Value Codec, an End Token or a Follow Token.
//
//  A Value codec (and its corresponding "introduce token") is a user defined callback function 
//  which gets "control" over the incoming character stream to decode data. In this meaning data 
//  can be hexadecimal values as well as strings or user defined data types. A Value codec is called 
//  (with an reset opcode) after recognising the corresponding token. Multiple Token can call this 
//  function. The external function decides when the end of the data is reached.
//
//  A follow token is a token which can occur multiple times after each other, but is treated as a
//  single occurance (can be used for space-characters)
//
//	An End token marks the end of the command and analyses the token buffer for a matching parse line
//  If one is found the corresponding function will be called.
//
//
//	Author 	: Rene Bllhoff
//  Created : around Jan. 2005
//
//	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
//
//////////////////////////////////////////////////////////////////////////////
//
//  This source file may be used and distributed without
//  restriction provided that this copyright statement is not
//  removed from the file and that any derivative work contains
//  the original copyright notice and the associated disclaimer.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//										
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public
//  License along with this source; if not, download it	from
//  http://www.gnu.org/licenses/gpl.html	
//
////////////////////////////////////////////////////////////////////////////

  #include "lib.h"


	//
  //  global datastruct for main control and parser codec 
	//
  
  typedef struct {
    UCHAR ucCurrentCharPointer;			// "virtual" pointer within all tokens (do reduce possibilites when searching tokens)
    UCHAR ucCurrentTokenPointer;		// write pointer within aucTokenBuffer Token Array. holds recognized tokens
    UCHAR ucCurrentValuePointer;		// write pointer within aucValueBuffer Value Array. holds values and parameters found in parse-line
    UCHAR ucLastToken;              // used for external parser codecs (when TK_UNGET, for dropping token) or follow token
    
    T_PARSE_FUNC pfCurrentParse;		// pointer to current parse-codec (if selected, otherwise NULL)
  
    UCHAR aucTokenBuffer    [SIZE_TOKEN_BUFFER];		// Token Buffer. Holds recognized Tokens. Cleared when token with TC_END found
    UCHAR aucValueBuffer    [SIZE_VALUE_BUFFER];		// Value Buffer. Holds values, which dropped by the codecs (e.g. Hex-Codec, String-Codec, Struct-Codec)
    UCHAR aucActiveSearches [(NUM_TOKENS / 8) + 1];	// Bit coded array signalling with each bit wether token is possible or not.
  } GLOBAL_DATA;

  //
  //  array of pointers to parser codec functions
  //
  
  #define TOKEN_CONTROL(control,func) (T_PARSE_FUNC) func,
    const T_PARSE_FUNC apfTokenFunctions [] = {
			NULL,   // fill up for TC_END
			NULL,   // fill up for TC_STD
			NULL,   // fill up for TC_FOLLOW
      #include PARSER_CONFIG_FILE
    };
  #undef TOKEN_CONTROL 

  
  //
  //  array of definitions of tokens
  //


  #define TOKEN(token,control,word) { (UCHAR) control, (UCHAR) sizeof (word) - 1, (UCHAR *) word },
    const TOKEN_DEF astTokenDefs [] = {
      #include PARSER_CONFIG_FILE
    };
  #undef TOKEN
  
  //
  //  parse-lines as array of bytes
  //

	#define PARSE_LINES  
		#define PARSE_BEGIN(parse, func)          const UCHAR aucParseElement_##parse [] =  { /* tokens treated as bytes */
		#define PARSE_DEF(value)                    value,
		#define PARSE_END                         };
			#include PARSER_CONFIG_FILE
		#undef PARSE_BEGIN
		#undef PARSE_DEF
		#undef PARSE_END
	#undef PARSE_LINES 
  
  //
  //  parse-lines as indexable structures (with pointer to raw bytes and execute function)
  //
  
	#define PARSE_LINES  
		#define PARSE_BEGIN(parse, func)         { (sizeof (aucParseElement_##parse) / sizeof (UCHAR)), (UCHAR *) (aucParseElement_##parse), (T_PARSE_EXEC) & func },
		#define PARSE_DEF(value)  
		#define PARSE_END       
			const PARSE_LINE astParseLines [] = {		/* generate entry for every parse-line entry */
				#include PARSER_CONFIG_FILE
			};
		#undef PARSE_BEGIN
		#undef PARSE_DEF
		#undef PARSE_END  
	#undef PARSE_LINES 
  
  
  //
  // token and control macros
  //

	// gets attribute of token
  #define ucGetTokenAttribute(x)   ( astTokenDefs [(x)].ucTokenAttribute )
	// gets pointer to token text
  #define ucGetTokenText(x)        ( astTokenDefs [(x)].pucTokenText )
  // gets pointer to parser execute function
  #define pfGetTokenFunction(x)    ( (T_PARSE_FUNC) (apfTokenFunctions [astTokenDefs [(x)].ucTokenAttribute]) )

	//
  // parse line macros
  //

	// returns length of parse line
  #define ucGetLengthParse(x)      ( astParseLines [x].ucNumberTokens )

  // global variables

  GLOBAL_DATA stGlobalParse;
  
  // static used functions
  
	// reduces possibilites of token with each incoming character 
  static UCHAR  ucReducePossibleSearches   (UCHAR ucCharPos, UCHAR ucCurrChar, UCHAR *pucTokenNumber);
	// recognition of token
  static UCHAR  ucProcessTokenChar         (UCHAR ucChar);
	// resets token recognition
  static VOID   vResetTokenFinder          (VOID);
	// clears token buffer
  static VOID   vClearTokenBuffer          (VOID);
	// clears value vuffer
  static VOID   vClearValueBuffer          (VOID);
	// resets all codecs
  static VOID   vResetExternalParsers      (VOID);
	// check input (after end-token (attribute : TC_END)) and execute callback
  static UCHAR  ucCheckParseLines          (VOID);
	// drops token (byte) on stack
  static VOID   vDropToken                 (UCHAR ucToken);
	// checks if current char can be ignored (e.g. ctrl chars)
  static BOOL   bIsValidChar               (UCHAR ucX);
  
  
// initializing all parser parts, such as ...

VOID vParserInit (VOID)
{
  vResetTokenFinder      ();
  vClearTokenBuffer      (); 
  vClearValueBuffer      (); 
  vResetExternalParsers  ();

	stGlobalParse.ucLastToken = 0;
}

// ... recognition of tokens, ...

VOID vResetTokenFinder (VOID)
{
  memset (stGlobalParse.aucActiveSearches,  0xff  , sizeof (stGlobalParse.aucActiveSearches));  
  stGlobalParse.ucCurrentCharPointer = 0;  
  stGlobalParse.pfCurrentParse = NULL;
}

// ... token buffer, ...

VOID vClearTokenBuffer (VOID)
{
  memset (stGlobalParse.aucTokenBuffer,      0     , sizeof (stGlobalParse.aucTokenBuffer));
  stGlobalParse.ucCurrentTokenPointer = 0;
}

// ... value buffer, ...

VOID vClearValueBuffer (VOID)
{
  memset (stGlobalParse.aucValueBuffer,      0     , sizeof (stGlobalParse.aucValueBuffer));
  stGlobalParse.ucCurrentValuePointer = 0;
}

// ... and external codecs
VOID vResetExternalParsers (VOID)
{
  UCHAR ucC;
  
  for (ucC = 0; ucC < NUM_TOKEN_CONTROLS ; ucC ++)
  {
    if (apfTokenFunctions [ucC] != NULL)
    {
      apfTokenFunctions [ucC] ( EC_RESET, 0, 0);
    }
  }
}



// function, checking wether an incoming char can be ignored
  
BOOL bIsValidChar (UCHAR ucX)
{
  BOOL bIsValid = 1;
	// trick : simply build the ignore list from configuration file to concatenated if condition ...
  #define TOKEN_IGNORE_CHARS
    #undef CHAR_IGNORE_SINGLE
    #undef CHAR_IGNORE_AREA0
    #undef CHAR_IGNORE_AREA
    if ( // (in the magic if)
      #define CHAR_IGNORE_SINGLE(x)  (ucX == x) ||
      #define CHAR_IGNORE_AREA0(y)   (              (ucX <= y)) ||    
      #define CHAR_IGNORE_AREA(x,y)  ((ucX >= x) && (ucX <= y)) ||    
        #include PARSER_CONFIG_FILE
      #undef CHAR_IGNORE_SINGLE
      #undef CHAR_IGNORE_AREA0
      #undef CHAR_IGNORE_AREA
    0 )  // ... and don't forget to "close" the condition in if statement
    { bIsValid = 0; }
  #undef TOKEN_IGNORE_CHARS
  return bIsValid;
}  

// process character on high-level

VOID vParserProcessChar (UCHAR ucChar)
{
  UCHAR ucResult;
  
	if (ucChar == TOKEN_RESET_CHAR)  // if reset-char found, reset the whole parser
  {
    vParserInit (); 
		return;
  }

  ucChar = ucUpperCase (ucChar); // if no reset-char, convert incoming character to uppercase (allows easier handling)
  
  if ((bIsValidChar (ucChar)))   // then filter by chars to ignore ...
  {
    ucResult = ucProcessTokenChar (ucChar);  // check for token
    if (! (((ucResult) == TK_NONE) || ((ucResult) == TK_SEARCHING)) )
    { 
			// token found ... check token attribute
      if (ucGetTokenAttribute (ucResult) == TC_END)  // end token ?
      {
        ucCheckParseLines ();			// process token buffer
        vParserInit ();						// initialize all to ensure proper function
      }
      else
      {
        vDropToken (ucResult);		// if token has std attribute, drop it in token buffer
      }
    }
    if (ucResult == TK_NONE)			// if no token found, reset ...
    {
      vParserInit ();
    }
  } 
}

// processes string on high level

VOID vParserProcessString (UCHAR *pucChar)
{
  while (*pucChar != 0)  // any comment ?!
  {
    vParserProcessChar (*pucChar);
    *pucChar++;
  }
}

// processes char on token-level

UCHAR ucProcessTokenChar (UCHAR ucChar)
{
  UCHAR ucNumPossib, // holds number of possible tokens after this char has been checked
				ucToken;		 // resulting token
  
  ucChar = ucUpperCase (ucChar);	// processes incoming char to uppercase
  
  // simple token or codec ?
  if (stGlobalParse.pfCurrentParse == NULL)
  {
		// token : get number of possible tokens after this char
    ucNumPossib = ucReducePossibleSearches (stGlobalParse.ucCurrentCharPointer, ucChar, &ucToken);
    // increment global "virutal" character pointer
		stGlobalParse.ucCurrentCharPointer++; 
    // if more than one possibility ... 
    if (ucNumPossib > 1)
    {
			// ... keep on searching
      return TK_SEARCHING;
    }
    // token found ? 
    else if (ucNumPossib == 1)
    {
			// if token really found ...
      if (ucToken != TK_SEARCHING)
      {
				// check wether it has the follow attribute ...
				if (astTokenDefs [ucToken].ucTokenAttribute == TC_FOLLOW)
				{
					if (stGlobalParse.ucLastToken == ucToken)	// if last token was the same -> keep on searching until difference
					{
						vResetTokenFinder ();  // to allow correct recognition of following tokens 
						return TK_SEARCHING;
					}
				}
				// ... otherwise we store it, and check if current token has a codec ...
				stGlobalParse.ucLastToken = ucToken;
				stGlobalParse.pfCurrentParse = pfGetTokenFunction (ucToken);
				if (stGlobalParse.pfCurrentParse == NULL) // ... 'cause if it is a regular token, we finish (by resetting the "machine" and return token found)
				{
					vResetTokenFinder ();
					return ucToken;
				}
				else  // .. otherwise we give control to the codec by "faking" that current token has not ended
				{
					return TK_SEARCHING;
				}
      }
      else  // if we're searching, we're searching !!
      {
        return TK_SEARCHING;
      }  
    }
    // no token found ->
    else if (ucNumPossib == 0)
    {
			// reset token recognition and report "no token"
      vResetTokenFinder ();
      return TK_NONE;
    }
	}
  else
  {
		// we call the codec ...
    ucToken = stGlobalParse.pfCurrentParse ( EC_PROCESS, ucChar, stGlobalParse.ucLastToken);
    switch (ucToken)
    {
			// and report that we're searching or found no token (in case of error)
      case TK_NONE  :
      case TK_SEARCHING :
        break;
      // drop token, and process current char (it could be a space or something)
      case TK_UNGET :
        vDropToken (stGlobalParse.ucLastToken);  // 
        vResetTokenFinder (); 
        vResetExternalParsers  ();
        return ucProcessTokenChar (ucChar);     
      default :
        vResetTokenFinder (); 
        vResetExternalParsers  ();
    }
    return ucToken;
  }
  return TK_NONE;
}

// helping function for recognition of tokens

UCHAR ucReducePossibleSearches (UCHAR ucCharPos, UCHAR ucCurrChar, UCHAR *pucTokenNumber)
{
  UCHAR ucC,            // zhler ber alle worte
        ucByte,         // byte innerhalb der aktiv liste
        ucResult = 0;   // anzahl der mglichkeiten die brigbleiben

  for (ucC = 0; ucC < NUM_TOKENS; ucC++)
  {
    ucByte = ucC >> 3;
    if ( ((stGlobalParse.aucActiveSearches [ucByte] & (1 << (ucC & 7))) != 0) &&
         //(strlen ((char *) (astTokenDefs [ucC].pucTokenText)) >= ucCharPos )
         ( (astTokenDefs [ucC].ucLength) >= ucCharPos )
       )
    {
      if (ucUpperCase (astTokenDefs [ucC].pucTokenText [ucCharPos]) == ucCurrChar)
      {
        ucResult++;
        
        //if ( (strlen ((char *) (astTokenDefs [ucC].pucTokenText)) == (ucCharPos + 1) ) )
        if ( (astTokenDefs [ucC].ucLength) == (ucCharPos + 1) )
        {
          *pucTokenNumber = ucC;
        }
        else
        {
          *pucTokenNumber = TK_SEARCHING;
        }  
      }
      else
      {
        // token-suche fr diesen eintrag zurcksetzen (soll zeit sparen)
        stGlobalParse.aucActiveSearches [ucByte] &= ~(1 << (ucC & 7));      
      }
    }
  }

  return ucResult;
}


// buffer operation functions 

// drop token on token buffer until (!) its full

VOID vDropToken (UCHAR ucToken)
{
	// simply drop on "stack"
  if ((ucToken < NUM_TOKENS) && (stGlobalParse.ucCurrentTokenPointer <= SIZE_TOKEN_BUFFER))
  {
    stGlobalParse.aucTokenBuffer [stGlobalParse.ucCurrentTokenPointer++] = ucToken; 
  }
}

// drop byte on value buffer

VOID vParserDropValue (UCHAR ucValue)
{
	// simply drop on "stack" until (!) its full
  if ( stGlobalParse.ucCurrentValuePointer <= SIZE_VALUE_BUFFER )
  {
    stGlobalParse.aucValueBuffer [stGlobalParse.ucCurrentValuePointer++] = ucValue; 
  }
}

// check token buffer for existing parse-line (simply compare, might be optimized) and execute it if found

UCHAR ucCheckParseLines (VOID)
{
  UCHAR ucC, ucD,
        ucFlag;

	// simply go thru all parse lines
  for (ucC = 0; ucC < NUM_PARSES; ucC ++)
  {
		// when checking be optimistic ....
    ucFlag = 1;
    for (ucD = 0; ucD < ucGetLengthParse(ucC); ucD ++)
    {    
			// every token might be the last for this parse line
      ucFlag &= (astParseLines [ucC].pstParseElements [ucD] == stGlobalParse.aucTokenBuffer [ucD]);
			if (ucFlag == FALSE) break; // we don't need to check all tokens in the parse line if any doesn't fit
    }
		// if found parse, execute it
    if (ucFlag == 1)
    {
			// call execute function
      astParseLines [ucC].pfCallFunction (ucC, stGlobalParse.aucValueBuffer);
			// reset all
      vResetTokenFinder      ();
      vClearTokenBuffer      ();
      vClearValueBuffer      ();
      vResetExternalParsers  ();
      // report parse number
      return ucC;
    }
  }  
  return NO_PARSE;
}

